
abstract class Enemy extends Collider {
    sl: ShmupLayer;
    hp = 1;
    xp = 1;
    particleColor = "#2e6a42";

    img: string | null = null;
    imgOffset: [number, number] = [0, 0];
    spawnTime: number;

    constructor(sl: ShmupLayer, pos: [number, number], rad: number) {
        super(pos, rad);
        this.sl = sl;
        this.spawnTime = sl.playTime;
    }

    abstract update(deltat: number): void;

    ko() {
        this.keep = false;
        this.sl.handleEnemyDestroyed(this);
    }

    handleCollideWithPlayer() {
        this.takeDamage(250);
    }

    takeDamage(amt: number) {
        if (this.hp > 0) {
            this.hp -= amt;
            if (this.hp <= 0) {
                this.ko();
            }
        }
    }

    abstract estVel(): [number, number];
}


class BasicEnemy extends Enemy {
    // A basic enemy that moves forward and shoots
    lastFiredTime: number;

    speed = 100;
    fireInterval = 1.5;
    hp = 15;
    xp = 15;
    bulletSpeed = 300;

    img = 'basicenemy';
    imgOffset: [number, number] = [15, 14];

    constructor(sl: ShmupLayer, pos: [number, number]) {
        super(sl, pos, 7);
        this.lastFiredTime = sl.playTime;
    }

    update(deltat: number) {
        // update position
        this.moveTo([this.pos[0] - this.speed * deltat, this.pos[1]]);

        // check if we should fire
        if (this.sl.playTime > this.lastFiredTime + this.fireInterval) {
            this.sl.projectiles.push(new Projectile([this.pos[0] - this.rad, this.pos[1]],
                [-this.bulletSpeed, 0], [Collision.Player], 1, this));
            this.lastFiredTime += this.fireInterval;
        }

        // check if we have moved out
        if (this.pos[0] < -EXISTS_RAD) {
            this.keep = false;
        }
    }

    estVel(): [number, number] {
        return [-this.speed, 0];
    }
}


class GhostEnemy extends BasicEnemy {
    ghost = true;
    particleColor = "#b45eb3";
    img = 'ghostenemy';
}


class KamikazeEnemy extends Enemy {
    // An enemy that flies forwards fast, possibly with a sinewave

    speed = 250;
    period = 1.8 / (2 * Math.PI);
    hp = 10;
    xp = 15;
    img = 'kamikazeenemy';
    imgOffset: [number, number] = [11, 9];

    yBase: number;
    sineAmp: number;
    spawnTime: number;

    constructor(sl: ShmupLayer, pos: [number, number], sineAmp: number) {
        super(sl, pos, 9);
        this.yBase = pos[1];
        this.sineAmp = sineAmp;
        this.spawnTime = this.sl.playTime;
    }

    update(deltat: number) {
        // update position

        let x = this.pos[0] - this.speed * deltat;
        let time = this.sl.playTime - this.spawnTime;
        let y = this.yBase + this.sineAmp * Math.sin(time / this.period);
        this.moveTo([x, y]);

        // check if we have moved out
        if (this.pos[0] < -EXISTS_RAD) {
            this.keep = false;
        }
    }

    estVel(): [number, number] {
        let time = this.sl.playTime - this.spawnTime;
        let xv = -this.speed;
        let yv = this.sineAmp * Math.cos(time / this.period) / this.period;
        return [xv, yv];
    }
}


class TrackerEnemy extends Enemy {
    // An enemy that flies forwards, shoots towards the player, then flies back away.

    destX: number;
    travTime = 0.6;
    stayTime = 1.2;
    shotSpeed = 300;
    hp = 10;
    xp = 20;
    img = 'trackerenemy';
    imgOffset: [number, number] = [9, 9];

    numShotsDone = 0;
    numShots: number;
    nextFireTime: number | null;

    constructor(sl: ShmupLayer, destPos: [number, number],
        numShots: number) {
        super(sl, [FLY_START, destPos[1]], 9);
        this.destX = destPos[0];
        this.numShots = numShots;
        if (numShots > 0) {
            this.nextFireTime = this.travTime + this.stayTime / (numShots + 1);
        }
        else {
            this.nextFireTime = null;
        }
    }

    shoot() {
        let x = this.pos[0];
        let y = this.pos[1];
        let destX = this.sl.pc.pos[0];
        let destY = this.sl.pc.pos[1];
        let dx = destX - x;
        let dy = destY - y;
        let norm = Math.sqrt(dx * dx + dy * dy);
        dx /= norm;
        dy /= norm;
        let offx = dx * this.rad;
        let offy = dy * this.rad;
        let vx = dx * this.shotSpeed;
        let vy = dy * this.shotSpeed;
        this.sl.projectiles.push(
            new Projectile([x + offx, y + offy],
                [vx, vy], [Collision.Player], 1, this));
    }

    getxxv() {
        let time = this.sl.playTime - this.spawnTime;
        let t1 = this.travTime;
        let t2 = t1 + this.stayTime;
        let t3 = t2 + this.travTime;
        let destX = this.destX;
        let x: number;
        let xv: number;
        if (time < t1) {
            let c = .5 * Math.PI / this.travTime;
            x = FLY_START + (destX - FLY_START) * Math.sin(time * c)
            xv = (destX - FLY_START) * Math.cos(time * c) * .5 * c;
        }
        else if (time < t2) {
            x = destX;
            xv = 0;
        }
        else if (time < t3) {
            let c = .5 * Math.PI / this.travTime;
            x = FLY_START + (destX - FLY_START) * Math.sin((t3 - time) * c);
            xv = (destX - FLY_START) * Math.cos((t3 - time) * c);
        }
        else {
            x = 0;
            xv = 0;
        }
        return [x, xv];
    }

    update(deltat: number) {
        let time = this.sl.playTime - this.spawnTime;
        let y = this.pos[1];

        let [x, xv] = this.getxxv();
        if (time > 2 * this.travTime + this.stayTime) {
            this.keep = false;
            return;
        }
        this.moveTo([x, y]);

        // Check if we have to fire
        if (this.nextFireTime !== null && time > this.nextFireTime) {
            this.shoot();
            this.numShotsDone += 1;
            if (this.numShotsDone < this.numShots) {
                this.nextFireTime = this.travTime + this.stayTime * (this.numShotsDone + 1) / (this.numShots + 1);
            }
            else {
                this.nextFireTime = null;
            }
        }
    }

    estVel(): [number, number] {
        let [x, xv] = this.getxxv();
        return [xv, 0];
    }
}

class MountainEnemy extends Enemy {
    // A basic enemy that moves forward and shoots
    lastFiredTime: number;

    speed = 40;
    fireInterval = 0.5;
    hp = 400;
    xp = 80;
    lastFiredAngle = -Math.PI / 2;
    img = 'mountainenemy';
    imgOffset: [number, number] = [16, 16];

    constructor(sl: ShmupLayer, pos: [number, number]) {
        super(sl, pos, 15);
        this.lastFiredTime = sl.playTime;
    }

    update(deltat: number) {
        // update position
        this.moveTo([this.pos[0] - this.speed * deltat, this.pos[1]]);

        // check if we should fire
        if (this.sl.playTime > this.lastFiredTime + this.fireInterval) {
            this.lastFiredAngle -= Math.PI / 6;
            let vx = Math.cos(this.lastFiredAngle);
            let vy = -Math.sin(this.lastFiredAngle);
            this.sl.projectiles.push(new Projectile(
                [this.pos[0] + 10 * vx, this.pos[1] + 10 * vy],
                [200 * vx, 200 * vy], [Collision.Player], 1, this));
            this.lastFiredTime += this.fireInterval;
        }

        // check if we have moved out
        if (this.pos[0] < -EXISTS_RAD) {
            this.keep = false;
        }
    }

    estVel(): [number, number] {
        return [-this.speed, 0];
    }
}

class BossEnemy extends Enemy {

    destX = 500;
    travTime = 2.0;
    maxhp = 2400;
    hp = 2400;
    xp = 0;
    mode = 0;
    changeModeTime = 0;

    modeTrav = 0;
    modeRest = 1;
    modeSpiral = 2;
    modeGhost = 3;
    modeTracking = 4;

    lastSpiralAttackTime = 0;
    lastSpiralAng = 0;
    spiralFactor = 1;
    lastTrackingAttackTime = 0;

    img = 'boss';
    imgOffset: [number, number] = [11, 15];
    particleColor = "#141020";

    constructor(sl: ShmupLayer) {
        super(sl, [FLY_START, 180], 11);
    }

    changeMode(n: number) {
        this.mode = n;
        this.changeModeTime = this.sl.playTime;
        if (this.mode === this.modeSpiral) {
            this.lastSpiralAttackTime = this.sl.playTime;
        }
        else if (this.mode === this.modeTracking) {
            this.lastTrackingAttackTime = this.sl.playTime;
        }
    }

    updateTravel() {
        let t = this.sl.playTime - this.spawnTime;
        if (t < this.travTime) {
            let a = t / this.travTime;
            this.moveTo([a * this.destX + (1 - a) * FLY_START, 180]);
        }
        else {
            this.moveTo([this.destX, 180]);
            this.changeMode(this.modeRest);
        }
    }

    updateRest() {
        let t = this.sl.playTime - this.changeModeTime;
        if (t > 2.0) {
            this.changeMode(this.modeSpiral);
        }
    }

    shootSpiral() {
        if (Math.random() < 0.002) {
            this.spiralFactor *= -1;
        }
        this.lastSpiralAng += this.spiralFactor * Math.PI * 2 / 3.03;
        let dx = Math.cos(this.lastSpiralAng);
        let dy = Math.sin(this.lastSpiralAng);
        let x = this.pos[0] + 15 * dx;
        let y = this.pos[1] + 15 * dy;
        let vx = 200 * dx;
        let vy = 200 * dy;
        this.sl.projectiles.push(new Projectile(
            [x, y],
            [vx, vy], [Collision.Player], 1, this));
    }

    updateSpiral() {
        let ts = 0.01;
        while (this.lastSpiralAttackTime < this.sl.playTime) {
            this.shootSpiral();
            this.lastSpiralAttackTime += ts;
        }
        if (this.sl.playTime - this.changeModeTime > 10.0) {
            this.changeMode(this.modeGhost);
        }
    }

    updateGhost() {
        this.ghost = true;
        if (this.sl.playTime - this.changeModeTime > 7.5) {
            this.ghost = false;
            this.changeMode(this.modeTracking);
        }
    }

    shootTracking() {
        let dx = this.sl.pc.pos[0] - this.pos[0];
        let dy = this.sl.pc.pos[1] - this.pos[1];
        let r = Math.sqrt(dx * dx + dy * dy);
        dx /= r;
        dy /= r;
        let x = this.pos[0] + dx * 15;
        let y = this.pos[1] + dy * 15;
        let vx = 200 * dx;
        let vy = 200 * dy;
        this.sl.projectiles.push(new Projectile(
            [x, y],
            [vx, vy], [Collision.Player], 1, this));
    }

    updateTracking() {
        let ts = 0.1;
        while (this.lastTrackingAttackTime < this.sl.playTime) {
            this.shootTracking();
            this.lastTrackingAttackTime += ts;
        }
        if (this.sl.playTime - this.changeModeTime > 10.0) {
            this.changeMode(this.modeRest);
        }
    }

    update() {
        this.sl.bossHealthFraction = this.hp / this.maxhp;
        if (this.mode === this.modeTrav) {
            this.updateTravel();
        }
        else if (this.mode === this.modeRest) {
            this.updateRest();
        }
        else if (this.mode === this.modeSpiral) {
            this.updateSpiral();
        }
        else if (this.mode === this.modeGhost) {
            this.updateGhost();
        }
        else if (this.mode === this.modeTracking) {
            this.updateTracking();
        }
    }

    takeDamage(amt: number) {
        this.hp -= amt;
        this.sl.bossHealthFraction = this.hp / this.maxhp;
        if (this.hp <= 0) {
            this.ko();
        }
    }

    estVel(): [number, number] {
        return [0, 0];
    }
}