class Game {

    data: { [key: string]: any };
    mainFont: PH.CanvasFont;
    ctx: CanvasRenderingContext2D;
    pixelationLayer: PH.PixelationLayer;

    soundPlayer: PH.SoundPlayer;

    layerManager = new PH.LayerManager();

    prepLayer: PrepLayer;
    msgBoxLayer: MsgBoxLayer;
    shmupLayer: ShmupLayer | null = null;

    spriteBoxFrame: PH.SpriteBox;
    spriteBoxButton: PH.SpriteBox;
    spriteBoxPressed: PH.SpriteBox;
    spriteBoxHighlight: PH.SpriteBox;
    parallaxTime = 0;
    gameMusic: AudioBufferSourceNode | null = null;

    playedBefore = false;

    constructor(data: { [key: string]: any }, audioContext: AudioContext,
        mainFont: PH.CanvasFont, pixelationLayer: PH.PixelationLayer) {
        this.data = data;
        this.soundPlayer = new PH.SoundPlayer(audioContext, {});
        this.mainFont = mainFont;
        this.pixelationLayer = pixelationLayer;
        this.ctx = pixelationLayer.srcCtx;
        this.msgBoxLayer = new MsgBoxLayer(this);

        this.spriteBoxFrame = new PH.SpriteBox(this.data['boxes'], 8, 0);
        this.spriteBoxButton = new PH.SpriteBox(this.data['boxes'], 8, 1);
        this.spriteBoxPressed = new PH.SpriteBox(this.data['boxes'], 8, 2);
        this.spriteBoxHighlight = new PH.SpriteBox(this.data['boxes'], 8, 3);

        this.layerManager.setupMouseListeners(pixelationLayer.destCtx.canvas);
        this.layerManager.setupKeyboardListeners(window);

        this.layerManager.setTopLayers(this.msgBoxLayer, pixelationLayer);
        this.layerManager.setMainLayers(new MenuLayer(this));

        this.prepLayer = new PrepLayer(this);
        // Start animation frames.
        let fm = new PH.FrameManager({
            frameCallback: (deltat) => this.frame(deltat)
        });
        fm.start();
    }

    startGameMusic() {
        this.gameMusic = this.soundPlayer.playSound(this.data['ld46music'], true);
    }

    stopGameMusic() {
        if(this.gameMusic !== null) {
            this.soundPlayer.stopSound(this.gameMusic);
            this.gameMusic = null;
        }
    }

    frame(deltat: number) {
        // Update step.
        this.layerManager.update(deltat);
        let dilation: number;
        if (this.shmupLayer !== null) {
            dilation = this.shmupLayer.getDilation();
        }
        else {
            dilation = 0;
        }
        this.parallaxTime += deltat * dilation;

        // Graphics step.
        PH.resizeCanvasToSizeOnScreen(this.pixelationLayer.destCtx.canvas);
        this.ctx.drawImage(this.data['stars'], 0, 0);

        let moon = <HTMLImageElement>this.data['moon'];
        let moonSpacing = 800;
        let moonOffset = Math.floor((this.parallaxTime * 2 + 100) % moonSpacing);
        this.ctx.drawImage(moon, -moonOffset, 50);
        this.ctx.drawImage(moon, moonSpacing - moonOffset, 50);

        let clouds = <HTMLImageElement>this.data['clouds'];
        let cloudOffset = Math.floor((this.parallaxTime * 50) % clouds.width);
        this.ctx.drawImage(clouds, -cloudOffset, 50);
        this.ctx.drawImage(clouds, clouds.width - cloudOffset, 50);

        this.layerManager.draw();
    }

    endMenu() {
        this.startPrep();
    }

    startPrep() {
        this.layerManager.setMainLayers(this.prepLayer, this.prepLayer.uiLayer, this.prepLayer.prepIconLayer);
        this.shmupLayer = null;
        this.playedBefore = true;
        this.stopGameMusic();
    }

    startShmup(skillChoices: SkillChoice[]) {
        let shmupLayer = new ShmupLayer(this, skillChoices);
        this.layerManager.setMainLayers(shmupLayer);
        this.shmupLayer = shmupLayer;
        this.startGameMusic();
    }

    startEnding() {
        let endingLayer = new EndingLayer(this);
        this.layerManager.setMainLayers(endingLayer);
        this.shmupLayer = null;
    }

    bindKey(skill: Skill): Promise<string> {
        let prom = new Promise<string>((resolve, reject) => {
            let bl = new BindingLayer(this, skill, (binding: string) => {
                this.layerManager.setMainLayers(this.prepLayer, this.prepLayer.uiLayer, this.prepLayer.prepIconLayer);
                resolve(binding);
            });
            this.layerManager.setMainLayers(this.prepLayer, this.prepLayer.uiLayer, this.prepLayer.prepIconLayer, bl);
        });
        return prom;
    }

    msgBox(msg: string): Promise<void> {
        let prom = new Promise<void>((resolve, reject) => {
            this.msgBoxLayer.display(msg, resolve);
        });
        return prom;
    }

}

async function start() {
    window.onerror = null;

    let mainGameCanvas = PH.createCanvas(640, 360);
    let outGameCanvas = <HTMLCanvasElement>document.getElementById('outGameCanvas')!;
    outGameCanvas.hidden = false;
    let audioContext = new AudioContext();
    let loader = new PH.Loader(audioContext);

    let ctx = mainGameCanvas.getContext('2d')!;
    let outCtx = outGameCanvas.getContext('2d')!;
    let pixelationLayer = new PH.PixelationLayer(ctx, outCtx);

    // Load a font
    let mainFont = <PH.PixelFont>await loader.getFile('m5x7.bff');
    mainFont.img = PH.changeImageColor(mainFont.img, [219, 224, 231]);
    mainFont.yOffset = -4;
    mainFont.lineHeight = 10;

    // Set up loading sceen
    let loadingScreen = new LoadingScreen(loader, ctx, mainFont);

    // Start animation frames for while the game is loading.
    let fm = new PH.FrameManager({
        frameCallback: (deltat) => {
            PH.resizeCanvasToSizeOnScreen(outCtx.canvas);
            PH.fillCanvas(ctx, "#0d2f6d");
            loadingScreen.draw();
            pixelationLayer.draw();
        }
    });
    fm.start();

    // Load the main contents of the game.
    let data = await loader.getFile('game.dat', (bytes, totalBytes) => loadingScreen.setProgress(bytes, totalBytes));

    fm.stop();
    new Game(data, audioContext, mainFont, pixelationLayer);
}

window.onload = start;
