class PrepIconLayer extends PH.Layer {
    // Draw the icons over the top of the skill buttons
    game: Game;
    prepLayer: PrepLayer;

    constructor(game: Game, prepLayer: PrepLayer) {
        super();
        this.game = game;
        this.prepLayer = prepLayer;
    }

    draw() {
        let skillButtonsLeft = this.prepLayer.skillButtonsLeft;
        let skillButtonsTop = this.prepLayer.skillButtonsTop;
        let skillButtonsXStep = this.prepLayer.skillButtonsXStep;
        let skillButtonsYStep = this.prepLayer.skillButtonsYStep;
        let skillButtonsPerRow = this.prepLayer.skillButtonsPerRow;
        let skills = this.prepLayer.skills;
        for (let k = 0; k < skills.length; k++) {
            let srcX = 32 * k;
            let destX = 4 + skillButtonsLeft + (k % skillButtonsPerRow) * skillButtonsXStep;
            let destY = 4 + skillButtonsTop + Math.floor(k / skillButtonsPerRow) * skillButtonsYStep;
            this.game.ctx.drawImage(this.game.data['icons'],
                srcX, 0, 32, 32, destX, destY, 32, 32);
        }
    }
}

class PrepLayer extends PH.Layer {
    game: Game;
    uiLayer: PH.CanvasUILayer;

    skillButtonsLeft = 16;
    skillButtonsTop = 96;
    skillButtonsPerRow = 6;
    skillButtonsXStep = 52;
    skillButtonsYStep = 52;
    skillButtonsWidth = 40;
    skillButtonsHeight = 40;

    descLeft = 16;
    descTop = 260;
    descWidth = 304;
    descHeight = 84;

    choiceLeft = 332;
    choiceTop = 36;
    choiceWidth = 288;
    choiceHeight = 44;
    choiceYStep = 48;
    bindingLeft = 526;
    bindingTop = 58;
    bindingWidth = 88;
    bindingHeight = 16;

    startLeft = 580;
    startTop = 328;
    startWidth = 40;
    startHeight = 16;

    skills: Skill[];
    choices: (SkillChoice | null)[] = [];
    hoverMsg = new Map<PH.CanvasButton, string[]>();
    selLevelIdx = 0;
    choiceRects: PH.Rect[] = [];
    mousePos: PH.MousePosition = null;
    bindingButtons: PH.CanvasButton[] = [];
    active = true;
    prepIconLayer: PrepIconLayer;

    constructor(game: Game) {
        super();
        this.game = game;

        // Specifies the skills, in the order that their icons are stored in the icons image.
        this.skills = [
            new Skills.MachineCannon(),
            new Skills.SprayCannon(),
            new Skills.QuantumCannon(),
            new Skills.HybridCannon(),
            new Skills.PowerBlast(),
            new Skills.ParticleBomb(),
            new Skills.BigBomb(),
            new Skills.GhostBomb(),
            new Skills.Stab(),
            new Skills.Shield(),
            new Skills.Deflect(),
            new Skills.Heal(),
            new Skills.TimeDilation(),
            new Skills.SpiritBoost(),
            new Skills.AbsorbSpirit(),
            new Skills.AbsorbMind(),
        ]

        this.uiLayer = new PH.CanvasUILayer();
        let buttonDrawer = new PH.CanvasButtonSpriteDrawer(
            game.spriteBoxButton,
            game.spriteBoxPressed,
            game.mainFont);

        for (let k = 0; k < this.skills.length; k++) {
            let skill = this.skills[k];
            let x = this.skillButtonsLeft + (k % this.skillButtonsPerRow) * this.skillButtonsXStep;
            let y = this.skillButtonsTop + Math.floor(k / this.skillButtonsPerRow) * this.skillButtonsYStep;

            let b = new PH.CanvasButton(game.ctx, x, y,
                this.skillButtonsWidth, this.skillButtonsHeight,
                (b) => { this.skillClicked(<Skill>b.tag); }, "",
                buttonDrawer, [0], skill);
            this.uiLayer.addButton(b);
            let ww = this.game.mainFont.wordWrap(skill.message, this.descWidth - 16);
            ww.splice(0, 0, skill.name, "");
            this.hoverMsg.set(b, ww);
        }

        for (let k = 0; k < NUM_LEVELS; k++) {
            this.choices.push(null);
            this.choiceRects.push(new PH.Rect(this.choiceLeft, this.choiceTop + this.choiceYStep * k,
                this.choiceWidth, this.choiceHeight));
            this.bindingButtons.push(new PH.CanvasButton(this.game.ctx,
                this.bindingLeft, this.bindingTop + this.choiceYStep * k,
                this.bindingWidth, this.bindingHeight,
                (b) => { this.bindingClicked(b.tag); }, "Set key binding",
                buttonDrawer, [0], k));
        }

        let b = new PH.CanvasButton(game.ctx, this.startLeft, this.startTop,
            this.startWidth, this.startHeight,
            (b) => { this.startClicked(); }, "Start!",
            buttonDrawer, [0]);
        this.uiLayer.addButton(b);

        this.prepIconLayer = new PrepIconLayer(game, this);

    }

    handleLayerAdded() {
        if (!this.game.playedBefore) {
            this.quickMsg('In this game, you will level up and earn new skills as you defeat enemies. You have to choose all the skills you will earn before starting the action.\nClick on a level on the right hand side of the screen, and then click on a skill on the left to assign it to that level.\nMost skills are activated by a keyboard key. After assigning a skill, click the "Set key binding" button to assign a key for it. (Some "passive" skills do not need a key assigned.)\nClick "Start!" when you are ready to begin your quest.');
        }
    }

    async quickMsg(msg: string) {
        this.active = false;
        await this.game.msgBox(msg);
        this.active = true;
    }

    async skillClicked(skill: Skill) {
        // Check if we have already assigned this skill to a slot
        for (let k = 0; k < this.choices.length; k++) {
            let c = this.choices[k];
            if (k !== this.selLevelIdx && c !== null && c.skill === skill) {
                this.choices[k] = null;
            }
        }

        // Set the skill
        let levIdx = this.selLevelIdx;
        this.choices[levIdx] = new SkillChoice(skill, null);

        // Decide whether to have a "set key binding" button
        let b = this.bindingButtons[levIdx];
        let foundIdx = this.uiLayer.buttons.indexOf(b);
        if (foundIdx >= 0 && skill.passive) {
            this.uiLayer.buttons.splice(foundIdx, 1);
        }
        else if (foundIdx < 0 && !skill.passive) {
            this.uiLayer.addButton(b);
        }
        if (levIdx < this.choices.length - 1) {
            this.selLevelIdx++;
        }
        this.game.soundPlayer.playSound(this.game.data['click'], false);
    }

    async startClicked() {
        // check if every level has a skill assigned
        let allSkillsAssigned = true;
        for (let k = 0; k < this.choices.length; k++) {
            if (this.choices[k] === null) allSkillsAssigned = false;
        }
        if (!allSkillsAssigned) {
            let msg = "Not all skills are assigned. Select a level on the right by clicking on it. Then click on an available skill on the left to assign it.";
            await this.quickMsg(msg);
            return;
        }
        // check if every non-passive skill has a key bound
        let bindingsAssigned = true;
        for (let k = 0; k < this.choices.length; k++) {
            let c = this.choices[k]!;
            if (!c.skill.passive && c.binding === null) {
                bindingsAssigned = false;
            }
        }
        if (!bindingsAssigned) {
            let msg = 'Not all skills have key bindings. Click on a "Set key binding" button to assign a key to a skill.';
            await this.quickMsg(msg);
            return;
        }

        // start the shmup
        this.game.startShmup(<SkillChoice[]>this.choices);
    }

    async bindingClicked(idx: number) {
        let sc = this.choices[idx]!;
        this.active = false;
        let result = await this.game.bindKey(sc.skill);
        for(let sc of this.choices) {
            if(sc !== null && sc.binding === result) {
                sc.binding = null;
            }
        }
        sc.binding = result;
        this.active = true;
        return false;
    }

    draw() {
        let ctx = this.game.ctx;
        let mainFont = this.game.mainFont;
        this.game.spriteBoxFrame.draw(ctx, 8, 8, 624, 344);

        // Draw the title
        ctx.drawImage(this.game.data['title'], 60, 24);

        // Draw available skills text - the buttons are drawn by the UI layer.
        mainFont.drawText(ctx, "Available Skills",
            this.skillButtonsLeft, this.skillButtonsTop - 16);

        // Draw the box for hover text
        if (this.active) {
            this.game.spriteBoxFrame.draw(ctx, this.descLeft, this.descTop, this.descWidth, this.descHeight);
            let b = this.uiLayer.mouseOverButton;
            if (b !== null) {
                let lines = this.hoverMsg.get(b);
                if (lines) {
                    mainFont.drawMultiLineText(ctx, lines, this.descLeft + 8, this.descTop + 8);
                }
            }
        }

        // Draw the chosen skills
        mainFont.drawText(ctx, "Chosen Skills",
            this.choiceLeft, this.choiceTop - 16);
        for (let k = 0; k < NUM_LEVELS; k++) {
            let l = this.choiceLeft;
            let t = this.choiceTop + this.choiceYStep * k;
            let sb: PH.SpriteBox;
            if (k === this.selLevelIdx) {
                sb = this.game.spriteBoxHighlight;
            }
            else {
                sb = this.game.spriteBoxFrame;
            }
            sb.draw(ctx, this.choiceRects[k]);
            let choiceText = `Level ${k + 1}:`;
            let choice = this.choices[k];
            if (choice !== null) {
                choiceText += ` ${choice.skill.name}.`;
            }
            mainFont.drawText(ctx, choiceText, l + 8, t + 8);

            let bindingText: string;
            if (choice === null) {
                bindingText = '';
            }
            else if (choice.skill.passive) {
                bindingText = '(Passive)';
            }
            else if (choice.binding === null) {
                bindingText = '(No key bound)';
            }
            else {
                bindingText = `Binding: ${choice.binding}.`;
            }
            mainFont.drawText(ctx, bindingText, l + 8, t + 24);
        }
    }

    handleMouseMove(mousePos: PH.MousePosition) {
        this.mousePos = mousePos;
    }

    handleClick(button: number) {
        if (this.mousePos !== null) {
            for (let k = 0; k < this.choiceRects.length; k++) {
                let r = this.choiceRects[k];
                if (r.contains(this.mousePos)) {
                    this.selLevelIdx = k;
                    return false;
                }
            }
        }
        return true;
    }
}