/// <reference collider.ts>

// radius of projectiles
let PROJ_RAD = 2;

// keep projectiles and enemies in memory if they're still within this distance of the screen.
let EXISTS_RAD = 20;

// Enemies fly in from this (off-screen) x-coordinate.
let FLY_START = 660;

// Maximum level you can reach
let NUM_LEVELS = 6;

// XP requirement per level (going from level 1 => 2, ..., 5 => 6)
let LEVEL_COSTS = [300, 400, 500, 600, 700];

// Maximum number of particles before we force-recycle them
let MAX_PARTICLES = 2000;

// How long a particle lasts
let PARTICLE_LIFETIME = 1.0;

// How long to display special messages
let SPECIAL_MESSAGE_TIME = 3.0;

// How long to keep showing the shmup after victory/defeat
let ENDING_DURATION = 3.0;

enum Collision {
    Player,
    Physical,
    Ghost
}


class PlayerChar extends Collider {
    vel: [number, number] = [0, 0];

    rad = 3;
    movingUp = false;
    movingDown = false;
    movingLeft = false;
    movingRight = false;

    maxHP = 3;
    hp = 3;
    mana = 100;
    maxMana = 100;
    manaRechargeRate = 1.5;
    vincibleTime: number | null = null;
    level = 1;
    xp = 0;
    lastLevelXP = 0;

    img = 'player';
    imgOffset = [8, 11];
    particleStyle = "#c42c36";

    constructor() {
        super([60, 180], 4);
    }
}


class Projectile extends Collider {
    vel: [number, number];
    colls: Collision[];
    damageAmt: number;
    owner: Collider;

    constructor(pos: [number, number], vel: [number, number],
        colls: Collision[], damage: number, owner: Collider) {
        super(pos, PROJ_RAD);
        this.vel = vel;
        this.colls = colls;
        this.damageAmt = damage;
        this.owner = owner;
    }

    update(deltat: number, l: number, t: number, r: number, b: number) {
        // update position
        this.moveTo([this.pos[0] + this.vel[0] * deltat,
        this.pos[1] + this.vel[1] * deltat]);
        let ar = EXISTS_RAD;
        if (this.pos[0] < l - ar && this.vel[0] <= 0 ||
            this.pos[1] < t - ar && this.vel[1] <= 0 ||
            this.pos[0] > r + ar && this.vel[0] >= 0 ||
            this.pos[1] > b + ar && this.vel[1] >= 0) this.keep = false;
    }
}

class Particle {
    pos: [number, number] = [0, 0];
    vel: [number, number] = [0, 0];
    color: string = "";
    tEnd: number = 0;
}


class Wave {
    sl: ShmupLayer;
    time: number;

    constructor(sl: ShmupLayer) {
        this.sl = sl;
        this.time = sl.playTime;
    }

    async wait(dur: number) {
        await this.sl.waitUntilTime(this.time + dur);
        this.time += dur;
    }

    async go() {
        // --- EASY WAVE ---
        this.sl.showSpecialMessage("Wave 1/5");
        Patterns.sineY(this.sl, 8.0, 2, 33, 20, 340, Patterns.basicEnemy);
        await this.wait(10);
        Patterns.sineY(this.sl, 8.0, 2, 33, 340, 20, Patterns.basicEnemy);
        await this.wait(10);
        
        Patterns.lineY(this.sl, 2.0, 6, 60, 60, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        Patterns.lineY(this.sl, 2.0, 6, 300, 300, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        Patterns.lineY(this.sl, 2.0, 6, 180, 180, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        await this.wait(3);
        Patterns.lineY(this.sl, 2.0, 6, 120, 120, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        Patterns.lineY(this.sl, 2.0, 6, 240, 240, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        await this.wait(6);

        Patterns.sineY(this.sl, 8.0, 2, 33, 20, 340, Patterns.basicEnemy);
        await this.wait(10);
        Patterns.sineY(this.sl, 8.0, 2, 33, 340, 20, Patterns.basicEnemy);
        await this.wait(15);
        


        // --- SLIGHTLY TOUGHER WAVE ---
        this.sl.showSpecialMessage("Wave 2/5");
        Patterns.lineY(this.sl, 2.0, 6, 60, 60, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        Patterns.lineY(this.sl, 2.0, 6, 300, 300, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        Patterns.lineY(this.sl, 2.0, 6, 180, 180, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        await this.wait(3);
        Patterns.lineY(this.sl, 2.0, 6, 120, 120, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        Patterns.lineY(this.sl, 2.0, 6, 240, 240, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        await this.wait(6);

        Patterns.mountainEnemy(this.sl, FLY_START, 180);
        await this.wait(6);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));
        await this.wait(6);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));
        await this.wait(6);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));
        await this.wait(6);

        Patterns.sineY(this.sl, 8.0, 2, 33, 20, 340, Patterns.basicEnemy);

        await this.wait(14);


        // --- INTRODUCING GHOSTS WAVE ---
        this.sl.showSpecialMessage("Wave 3/5");
        Patterns.lineY(this.sl, 2.0, 6, 60, 60, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        Patterns.lineY(this.sl, 2.0, 6, 300, 300, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        Patterns.lineY(this.sl, 2.0, 6, 180, 180, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));

        await this.wait(4);
        Patterns.sineY(this.sl, 2.0, 0.5, 9, 20, 340, Patterns.basicEnemy);
        await this.wait(3);
        Patterns.sineY(this.sl, 2.0, 0.5, 9, 20, 340, Patterns.basicEnemy);
        await this.wait(3);

        Patterns.ghostEnemy(this.sl, FLY_START, 180);
        await this.wait(5);

        Patterns.sineY(this.sl, 2.0, 0.5, 9, 20, 340, Patterns.basicEnemy);
        await this.wait(3);
        Patterns.sineY(this.sl, 2.0, 0.5, 9, 340, 20, Patterns.basicEnemy);
        await this.wait(3);
        
        Patterns.ghostEnemy(this.sl, FLY_START, 180);
        await this.wait(0.5);
        Patterns.mountainEnemy(this.sl, FLY_START, 180);
        await this.wait(1);
        Patterns.ghostEnemy(this.sl, FLY_START, 180);
        await this.wait(0.5);
        Patterns.ghostEnemy(this.sl, FLY_START, 180);
        await this.wait(0.5);
        Patterns.ghostEnemy(this.sl, FLY_START, 180);
        await this.wait(0.5);
        Patterns.ghostEnemy(this.sl, FLY_START, 180);
        await this.wait(0.5);
        Patterns.ghostEnemy(this.sl, FLY_START, 180);
        await this.wait(0.5);

        Patterns.sineY(this.sl, 2.0, 0.5, 9, 20, 340, Patterns.basicEnemy);
        await this.wait(3);
        Patterns.sineY(this.sl, 2.0, 0.5, 9, 20, 340, Patterns.ghostEnemy);
        await this.wait(8);

        Patterns.mountainEnemy(this.sl, FLY_START, 180);
        await this.wait(2);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));
        await this.wait(2);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));
        await this.wait(2);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));
        await this.wait(2);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));

        await this.wait(10);


        // --- MOSTLY GHOSTS WAVE ---
        this.sl.showSpecialMessage("Wave 4/5");
        Patterns.sineY(this.sl, 2.0, 0.5, 9, 20, 340, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.sineY(this.sl, 2.0, 0.5, 9, 340, 20, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.sineY(this.sl, 2.0, 0.5, 9, 20, 340, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.sineY(this.sl, 2.0, 0.5, 9, 340, 20, Patterns.ghostEnemy);
        await this.wait(3);

        Patterns.lineY(this.sl, 0, 9, 20, 340, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.lineY(this.sl, 0, 8, 30, 330, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.lineY(this.sl, 0, 9, 20, 340, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.lineY(this.sl, 0, 8, 30, 330, Patterns.ghostEnemy);
        await this.wait(5);
        
        Patterns.lineY(this.sl, 0, 9, 20, 340, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.lineY(this.sl, 0, 8, 30, 330, Patterns.basicEnemy);
        await this.wait(3);
        Patterns.lineY(this.sl, 0, 9, 20, 340, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.lineY(this.sl, 0, 8, 30, 330, Patterns.basicEnemy);
        await this.wait(3);
        Patterns.lineY(this.sl, 0, 9, 20, 340, Patterns.ghostEnemy);
        
        await this.wait(10);


        // --- TOUGH WAVE ---
        this.sl.showSpecialMessage("Wave 5/5");
        Patterns.lineY(this.sl, 2.0, 6, 60, 60, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        Patterns.lineY(this.sl, 2.0, 6, 300, 300, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));
        Patterns.lineY(this.sl, 2.0, 6, 180, 180, (sl, x, y) => Patterns.kamikazeEnemy(sl, x, y, 40));

        await this.wait(4);
        Patterns.lineY(this.sl, 0, 4, 80, 280, Patterns.mountainEnemy);

        await this.wait(10);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));
        await this.wait(2);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));
        await this.wait(2);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));
        await this.wait(2);
        Patterns.lineY(this.sl, 2.0, 4, 80, 280, (sl, x, y) => Patterns.trackerEnemy(sl, 560, y, 5));

        await this.wait(8);
        Patterns.lineY(this.sl, 0, 9, 20, 340, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.lineY(this.sl, 0, 8, 30, 330, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.lineY(this.sl, 0, 9, 20, 340, Patterns.ghostEnemy);
        await this.wait(3);
        Patterns.lineY(this.sl, 0, 8, 30, 330, Patterns.ghostEnemy);

        await this.wait(10);

        // BOSS
        Patterns.bossEnemy(this.sl);
    }
}


class Waiter {
    keep = true;
    t: number;
    resolve: () => void;

    constructor(t: number, resolve: () => void) {
        this.t = t;
        this.resolve = resolve;
    }
}


class DamageCircle {
    pos: [number, number];
    rad: number;
    startTime: number;
    endTime: number;

    constructor(pos: [number, number], rad: number,
        startTime: number, endTime: number) {
        this.pos = pos;
        this.rad = rad;
        this.startTime = startTime;
        this.endTime = endTime;
    }
}


class SkillChoice {
    skill: Skill;
    binding: string | null;

    constructor(skill: Skill, binding: string | null) {
        this.skill = skill;
        this.binding = binding;
    }
}


class ShmupLayer extends PH.Layer {

    game: Game;

    rect: PH.Rect = new PH.Rect(0, 0, 640, 360);
    pc = new PlayerChar;
    enemies: Enemy[] = [];
    projectiles: Projectile[] = [];
    playTime = 0;
    waiters: Waiter[] = [];
    keysDown: Set<string> = new Set();
    skillChoices: SkillChoice[];
    curSkillBindings: { [key: string]: Skill } = {};
    curSkills: Skill[] = [];
    curCannonAttacks: Skill[] = [];
    nextCannonShotTime = 0;
    manaBonus = 0;
    xpBonus = 0;
    enemyDestroyedCallbacks: (() => void)[] = [];
    levelUpMessage: string[] = [];
    levelUpTime: number | null = null;
    particles: Particle[] = [];
    nextParticleIdx = 0;
    defeatTime: number | null = null;
    victoryTime: number | null = null;
    lastPlayerParticleTime = 0;
    lastCannonTing = 0;
    lastExplosionSound = 0;
    lastManaMsgTime: number | null = null;

    damageCircles: DamageCircle[] = [];
    flashTime: number | null = null;
    skillTimeDilation = 1;
    bossHealthFraction: number | null = null;
    specialMessageStartTime = 0;
    specialMessageText = "";

    constructor(game: Game, skillChoices: SkillChoice[]) {
        super();
        this.game = game;
        this.skillChoices = skillChoices;

    }

    async handleLayerAdded() {
        this.particles = [];
        for (let k = 0; k < MAX_PARTICLES; k++) {
            this.particles.push(new Particle());
        }
        this.lastPlayerParticleTime = this.playTime;

        // Uncomment this stuff when ready
        this.showSpecialMessage("Use arrow keys to move");
        await this.waitUntilTime(SPECIAL_MESSAGE_TIME);
        this.gainSkill(1);
        this.showLevelUp();
        await this.waitUntilTime(SPECIAL_MESSAGE_TIME + 2);

        (new Wave(this)).go();
    }

    showSpecialMessage(msg: string) {
        this.specialMessageStartTime = this.playTime;
        this.specialMessageText = msg;
    }

    getGameArea() {
        let l = this.rect.l
        let t = this.rect.t;
        let r = l + this.rect.w;
        let b = t + this.rect.h;
        return [l, t, r, b];
    }

    waitUntilTime(t: number): Promise<void> {
        let prom = new Promise<void>((resolve, reject) => {
            this.waiters.push(new Waiter(t, resolve));
        })
        return prom;
    }

    updatePlayer(deltat: number, l: number, t: number, r: number, b: number) {
        let pc = this.pc;

        // Check if the player returns from invincibility
        if (pc.vincibleTime !== null && this.playTime > pc.vincibleTime) {
            pc.vincibleTime = null;
        }

        // Apply acceleration from player's keyboard input
        let pcAcc = 960; // Acceleration due to player input
        let pcDec = 960; // Deceleration due to air resistance
        let pcMaxSpeed = 180;
        if (this.pc.movingUp) {
            pc.vel[1] -= pcAcc * deltat;
        }
        else if (pc.vel[1] < 0) {
            pc.vel[1] = Math.min(0, pc.vel[1] + pcDec * deltat);
        }
        if (this.pc.movingDown) {
            pc.vel[1] += pcAcc * deltat;
        }
        else if (pc.vel[1] > 0) {
            pc.vel[1] = Math.max(0, pc.vel[1] - pcDec * deltat);
        }
        if (this.pc.movingLeft) {
            pc.vel[0] -= pcAcc * deltat;
        }
        else if (pc.vel[0] < 0) {
            pc.vel[0] = Math.min(0, pc.vel[0] + pcDec * deltat);
        }
        if (this.pc.movingRight) {
            pc.vel[0] += pcAcc * deltat;
        }
        else if (pc.vel[0] > 0) {
            pc.vel[0] = Math.max(0, pc.vel[0] - pcDec * deltat);
        }

        // Slow down the speed - in case the speed is over the limit or due to
        // air resistance.
        let speed = Math.sqrt(pc.vel[0] * pc.vel[0] + pc.vel[1] * pc.vel[1]);
        let rat = 1;
        if (speed > pcMaxSpeed) {
            rat = pcMaxSpeed / speed;
        }
        pc.vel[0] *= rat;
        pc.vel[1] *= rat;

        // Update the player position based on velocity
        pc.moveTo([pc.pos[0] += pc.vel[0] * deltat, pc.pos[1] += pc.vel[1] * deltat]);
        if (pc.pos[0] < l + pc.rad) pc.pos[0] = l + pc.rad;
        if (pc.pos[1] < t + pc.rad) pc.pos[1] = t + pc.rad;
        if (pc.pos[0] > r - pc.rad) pc.pos[0] = r - pc.rad;
        if (pc.pos[1] > b - pc.rad) pc.pos[1] = b - pc.rad;

        // Update the player's mana
        if (pc.mana < pc.maxMana) {
            let deltaMana = pc.manaRechargeRate * deltat * (1 + this.manaBonus);
            this.gainMana(deltaMana);
        }

        // Create motion particles
        while (this.lastPlayerParticleTime < this.playTime) {
            let ts = Math.max(0.05, Math.min(0.2, 0.1 - pc.vel[0] / 300));
            this.lastPlayerParticleTime += ts;
            for (let k = 0; k < 3; k++) {
                let xOff = Math.random() * 4 - 2;
                let yOff = Math.random() * 4 - 2;
                let vxOff = Math.random() * 40;
                this.addParticle(pc.pos[0] - pc.rad - 2 + xOff, pc.pos[1] + yOff,
                    -200 - vxOff, 0, pc.particleStyle);
            }
        }
    }

    getDilation(): number {
        let d = 1;
        // Apply time dilation from skill
        d *= this.skillTimeDilation;

        // Apply time dilation from level up effect
        let luea = this.getLevelUpEffectAmount();
        d *= (1 - 0.85 * luea * luea);
        return d;
    }

    update(deltat: number) {
        let pc = this.pc;

        // Get the game area
        let [l, t, r, b] = this.getGameArea();

        deltat *= this.getDilation();

        // Update the play time
        this.playTime += deltat;

        // Update the player
        if (this.pc.hp > 0) {
            this.updatePlayer(deltat, l, t, r, b);
        }

        // Check if we've won the game
        if (this.bossHealthFraction !== null &&
            this.bossHealthFraction <= 0 &&
            this.victoryTime === null) {
            this.victory();
        }

        // Update enemies
        for (let e of this.enemies) {
            if (e.keep) {
                e.update(deltat);

                // check collision with player
                if (this.canDamagePlayer() && checkCollision(e, pc)) {
                    this.damagePlayer();
                    e.handleCollideWithPlayer();
                }
            }
        }
        this.enemies = this.enemies.filter(e => e.keep);

        // Update projectiles
        let ar = EXISTS_RAD;
        let pcpd = PROJ_RAD + pc.rad; // radius for checking collisions with the player
        for (let p of this.projectiles) {
            p.update(deltat, l, t, r, b);

            // check collision with player
            if (p.colls.indexOf(Collision.Player) >= 0 && this.canDamagePlayer() &&
                checkCollision(p, pc)) {
                this.damagePlayer();
                p.keep = false;
            }

            // check collision with physical & ghost enemies
            let pphys = (p.colls.indexOf(Collision.Physical) > -1);
            let pghost = (p.colls.indexOf(Collision.Ghost) > -1);
            if (pphys || pghost) {
                for (var e of this.enemies) {
                    // check if the attack suits the enemy
                    let works = (pphys && !e.ghost || pghost && e.ghost);
                    // now we actually check the collision
                    if (works && checkCollision(p, e)) {
                        e.takeDamage(p.damageAmt);
                        p.keep = false;
                        let ang = Math.random() * 2 * Math.PI;
                        let r = Math.random() * 50 + 100;
                        let vx = Math.cos(ang) * r;
                        let vy = Math.sin(ang) * r;
                        let c: string;
                        if (e.ghost) c = "#b45eb3";
                        else c = "#f7ac37";
                        this.addParticle(p.pos[0], p.pos[1], vx, vy, c);
                    }
                }
            }
        }
        this.projectiles = this.projectiles.filter(p => p.keep);

        // update particles
        for (let p of this.particles) {
            p.pos[0] += p.vel[0] * deltat;
            p.pos[1] += p.vel[1] * deltat;
        }

        // Check on any skills that have to run
        if (this.pc.hp > 0) {
            for (let s of this.curSkills) {
                s.update(deltat);
            };
        }

        // Check on any waiters that have to run
        for (let w of this.waiters) {
            if (this.playTime >= w.t) {
                w.resolve();
                w.keep = false;
            }
        }
        this.waiters = this.waiters.filter(w => w.keep);

        return true;
    }

    gainMana(deltaMana: number) {
        let pc = this.pc;
        pc.mana = pc.mana + deltaMana;
        pc.mana = Math.min(pc.mana, pc.maxMana);
        pc.mana = Math.max(pc.mana, 0);
    }

    canDamagePlayer(): boolean {
        return this.pc.hp > 0 && this.pc.vincibleTime == null &&
            this.victoryTime === null;
    }

    damagePlayer() {
        let pc = this.pc;
        pc.hp -= 1;
        if (pc.hp > 0) {
            this.makePlayerInvincible(2.0);
            this.game.soundPlayer.playSound(this.game.data['hitself'], false);
            
            this.addParticleExplosion(pc.pos[0], pc.pos[1],
                0, 0, pc.rad, 25, pc.particleStyle);
        }
        else if (pc.hp === 0) {
            this.defeat();
            this.explosionSound();
        }
    }

    async defeat() {
        let pc = this.pc;
        this.addParticleExplosion(pc.pos[0], pc.pos[1],
            pc.vel[0], pc.vel[1], pc.rad * 2, 250, pc.particleStyle);
        this.defeatTime = this.playTime;
        await this.waitUntilTime(this.playTime + ENDING_DURATION);
        this.game.startPrep();
    }

    async victory() {
        this.victoryTime = this.playTime;
        await this.waitUntilTime(this.playTime + ENDING_DURATION);
        this.game.startEnding();
    }

    makePlayerInvincible(dur: number) {
        let newT = this.playTime + dur;
        if (this.pc.vincibleTime === null || this.pc.vincibleTime < newT) {
            this.pc.vincibleTime = newT;
        }
    }

    damageEnemiesInRadius(pos: [number, number], rad: number, amt: number) {
        let rsq = rad * rad;
        for (let e of this.enemies) {
            let dx = pos[0] - e.pos[0];
            let dy = pos[1] - e.pos[1];
            let dsq = dx * dx + dy * dy;
            if (dsq < rsq) {
                e.takeDamage(amt);
            }
        }
    }

    deflectProjectiles(pos: [number, number], rad: number) {
        // Convert projectiles into player's projectiles,
        // and send them back towards their owners.
        let pc = this.pc;
        let radsq = rad * rad;
        for (let p of this.projectiles) {
            if (p.owner !== pc) {
                let pdx = (p.pos[0] - pc.pos[0]);
                let pdy = (p.pos[1] - pc.pos[1]);
                let pdsq = pdx * pdx + pdy * pdy;
                if (pdsq > radsq) continue;

                let oldOwner = p.owner;
                p.owner = this.pc;
                p.damageAmt = 2.0;
                p.colls = [oldOwner.ghost ? Collision.Ghost : Collision.Physical];

                let dx = oldOwner.pos[0] - p.pos[0];
                let dy = oldOwner.pos[1] - p.pos[1];
                let speed = p.vel[0] * p.vel[0] + p.vel[1] * p.vel[1];
                let curR = dx * dx + dy * dy;
                let ratio = Math.sqrt(speed / curR);
                dx *= ratio;
                dy *= ratio;
                p.vel = [dx, dy];
            }
        }
    }

    trySpendMana(amt: number) {
        if (this.pc.mana > amt) {
            this.pc.mana -= amt;
            return true;
        }
        else {
            this.showManaMsg();
            return false;
        }
    }

    handleEnemyDestroyed(e: Enemy) {
        this.addXP(e.xp * (1 + this.xpBonus));
        for (let f of this.enemyDestroyedCallbacks) {
            f();
        }

        let v = e.estVel();
        this.addParticleExplosion(e.pos[0], e.pos[1],
            v[0], v[1], e.rad, 100, e.particleColor);

        this.explosionSound();
    }

    addXP(xp: number) {
        // Add XP
        this.pc.xp += xp;
        
        // Check for a level up
        while (true) {
            let req = this.getNextLevelXP();
            if (req === null || this.pc.xp < req) break;
            this.pc.level += 1;
            this.pc.lastLevelXP = req;
            this.gainSkill(this.pc.level);
            this.showLevelUp();
        }
    }

    gainSkill(level: number) {
        // Gain the skill that was chosen for this level
        let skillChoice = this.skillChoices[level - 1];

        this.curSkills.push(skillChoice.skill);
        if (skillChoice.binding !== null) {
            this.curSkillBindings[skillChoice.binding] = skillChoice.skill;
        }
        skillChoice.skill.start(this);
    }

    getLevelUpEffectAmount() {
        if (this.levelUpTime === null) {
            return 0;
        }
        let a = (this.playTime - this.levelUpTime) / 2;
        if (a > 1) {
            this.levelUpTime = null;
            this.levelUpMessage = [];
            return 0;
        }
        if (a < 0.05) return a / 0.05;
        else if (a < 0.4) return 1;
        else return (1 - (a - 0.4) / 0.6)
    }

    showLevelUp() {
        this.levelUpTime = this.playTime;
        let sc = this.skillChoices[this.pc.level - 1];
        let bindingMessage: string;
        if (sc.binding === null) {
            bindingMessage = "(Passive Skill)";
        }
        else {
            bindingMessage = "Binding: " + sc.binding;
        }
        this.levelUpMessage = [`Level ${this.pc.level}`,
        `New skill: ${sc.skill.name}`, bindingMessage];
        setTimeout(() => {
            this.game.soundPlayer.playSound(this.game.data['levelup'], false);
        }, 100);
    }

    getNextLevelXP() {
        // compute the XP at the next level up
        let l = this.pc.level;
        if (l > LEVEL_COSTS.length) return null;
        let total = 0;
        for (let k = 0; k < l; k++) {
            total += LEVEL_COSTS[k];
        }
        return total;
    }

    addParticle(x: number, y: number, vx: number, vy: number, color: string) {
        let p = this.particles[this.nextParticleIdx];
        p.pos[0] = x;
        p.pos[1] = y;
        p.vel[0] = vx;
        p.vel[1] = vy;
        p.color = color;
        p.tEnd = this.playTime + PARTICLE_LIFETIME;

        this.nextParticleIdx = (this.nextParticleIdx + 1) % MAX_PARTICLES;
    }

    async addParticleExplosion(x: number, y: number,
        vx: number, vy: number, rad: number, count: number, color: string) {
        let t = this.playTime;
        let ts = 0.001;

        for (let k = 0; k < count; k++) {
            let curR = Math.random() * rad;
            let ang = Math.random() * 2 * Math.PI;
            let dx = Math.cos(ang);
            let dy = Math.sin(ang);
            this.addParticle(x + curR * dx, y + curR * dy,
                vx + 3 * curR * dx, vy + 3 * curR * dy, color);
            t += ts;
            if (t > this.playTime) {
                await this.waitUntilTime(ts);
            }
        }
    }

    explosionSound() {
        let ts = 0.1;
        if (this.playTime > this.lastExplosionSound + ts) {
            this.game.soundPlayer.playSound(this.game.data['explode'], false);
            this.lastExplosionSound = this.playTime;
        }
    }

    cannonTing() {
        let ts = 0.1;
        if (this.playTime > this.lastCannonTing + ts) {
            this.game.soundPlayer.playSound(this.game.data['cannon'], false);
            this.lastCannonTing = this.playTime;
        }
    }

    showManaMsg() {
        this.lastManaMsgTime = this.playTime;
    }

    draw() {
        let ctx = this.game.ctx;
        let pc = this.pc;

        // Draw damage circles
        ctx.fillStyle = "#ffffff";
        for (let c of this.damageCircles) {
            let a = 1 - (this.playTime - c.startTime) / (c.endTime - c.startTime);
            a = Math.max(a, 0);
            a = Math.min(a, 1);
            ctx.globalAlpha = a * 0.25;
            ctx.beginPath();
            ctx.arc(c.pos[0], c.pos[1], c.rad, 0, 2 * Math.PI);
            ctx.fill();
        }
        ctx.globalAlpha = 1;
        this.damageCircles = this.damageCircles.filter(c => this.playTime < c.endTime);

        // Draw enemies
        for (let e of this.enemies) {
            ctx.fillStyle = e.particleColor;
            if (e.ghost) {
                ctx.globalAlpha = 0.4 + 0.2 * Math.cos((this.playTime - e.spawnTime) * 0.7 * 2 * Math.PI);
            }
            if (e.img !== null) {
                let x = Math.floor(e.pos[0] - e.imgOffset[0]);
                let y = Math.floor(e.pos[1] - e.imgOffset[1]);
                ctx.drawImage(this.game.data[e.img], x, y);
            }
            else {
                ctx.fillRect(Math.floor(e.pos[0] - e.rad), Math.floor(e.pos[1] - e.rad),
                    e.rad * 2, e.rad * 2);
            }
            ctx.globalAlpha = 1;
        }

        // Draw the player position
        if (pc.hp > 0) {
            let pr = this.pc.rad;
            // ctx.fillStyle = pc.particleStyle;
            // ctx.fillRect(Math.floor(pc.pos[0] - pr), Math.floor(pc.pos[1] - pr),
            //     Math.floor(pr * 2), Math.floor(pr * 2));
            let x = Math.floor(pc.pos[0] - pc.imgOffset[0]);
            let y = Math.floor(pc.pos[1] - pc.imgOffset[1]);
            ctx.drawImage(this.game.data[pc.img], x, y);
            if (pc.vincibleTime !== null) {
                ctx.drawImage(this.game.data['shield'], x, y);
            }
            ctx.globalAlpha = 1;
        }

        // Draw particles
        for (let p of this.particles) {
            if (this.playTime < p.tEnd) {
                ctx.fillStyle = p.color;
                ctx.fillRect(Math.floor(p.pos[0]), Math.floor(p.pos[1]), 1, 1);
            }
        }

        // Draw all projectiles
        ctx.fillStyle = "#fbdf6b"
        let pImg = this.game.data['projectiles'];
        let pSize = 5;
        for (let p of this.projectiles) {
            let pIdx: number;
            let phys = p.colls.indexOf(Collision.Physical) >= 0;
            let ghost = p.colls.indexOf(Collision.Ghost) >= 0;
            if (p.owner !== pc) {
                pIdx = 3;
            }
            else if (phys && ghost) {
                pIdx = 2;
            }
            else if (ghost) {
                pIdx = 1;
            }
            else {
                pIdx = 0;
            }
            let x = Math.floor(p.pos[0] - pSize / 2);
            let y = Math.floor(p.pos[1] - pSize / 2);
            ctx.drawImage(pImg, pSize * pIdx, 0, pSize, pSize, x, y, pSize, pSize);
            //ctx.fillRect(Math.floor(p.pos[0] - 2), Math.floor(p.pos[1] - 2), 4, 4);
        }

        // Draw bomb flash
        if (this.flashTime !== null && this.playTime > this.flashTime) {
            let a = 1 - (this.playTime - this.flashTime) / 0.25;
            if (a < 0) {
                this.flashTime = null;
            }
            else {
                ctx.globalAlpha = a;
                ctx.fillStyle = "#ffffff";
                ctx.fillRect(0, 0, 640, 360);
                ctx.globalAlpha = 1;
            }
        }

        // Draw health
        this.game.mainFont.drawText(ctx, "HP", 4, 4);
        ctx.fillStyle = "#c42c36";
        for (let k = 0; k < this.pc.hp; k++) {
            let x = 34 + 20 * k;
            ctx.fillRect(x, 5, 16, 4);
        }

        // draw mana
        this.game.mainFont.drawText(ctx, "Mana", 4, 14);
        ctx.fillStyle = "#141020";
        ctx.fillRect(34, 16, 56, 4);
        ctx.fillStyle = "#13b2f2";
        ctx.fillRect(34, 16, Math.floor(this.pc.mana * 56 / this.pc.maxMana), 4);

        // draw xp
        this.game.mainFont.drawText(ctx, "XP", 4, 24);
        let nextXP = this.getNextLevelXP();
        let xpRatio: number;
        if (nextXP === null) {
            ctx.fillStyle = "#dbe0e7";
            xpRatio = 1;
        }
        else {
            ctx.fillStyle = "#141020";
            ctx.fillRect(34, 26, 56, 4);
            ctx.fillStyle = "#fbdf6b";
            xpRatio = (this.pc.xp - this.pc.lastLevelXP) / (nextXP - this.pc.lastLevelXP);
        }
        ctx.fillRect(34, 26, Math.max(1, Math.floor(xpRatio * 56)), 4);

        // draw boss health
        if (this.bossHealthFraction !== null) {
            let w = 536;
            ctx.fillStyle = "#141020";
            ctx.fillRect(100, 4, w, 20);
            ctx.fillStyle = "#2e6a42";
            let hw = Math.round(this.bossHealthFraction * w);
            ctx.fillRect(100, 4, hw, 20);
            this.game.mainFont.drawText(ctx, "EXCEPTION MONARCH", 104, 10);
            ctx.strokeStyle = "#141020"
            ctx.strokeRect(100.5, 4.5, hw, 20);
        }

        // Draw mana message
        if (this.lastManaMsgTime !== null &&
            this.playTime < this.lastManaMsgTime + 2.0) {
            let x = pc.pos[0];
            let y = pc.pos[1] - 20;
            this.game.mainFont.drawCenteredText(ctx, "LOW MANA", x, y);
        }

        // Draw initial message
        let timeSinceSpecialMessage = this.playTime - this.specialMessageStartTime;
        if (timeSinceSpecialMessage < SPECIAL_MESSAGE_TIME) {
            let a = timeSinceSpecialMessage / SPECIAL_MESSAGE_TIME;
            let amt: number;
            if (a < 0.1) {
                amt = a / 0.1;
            }
            else if (a < 0.9) {
                amt = 1;
            }
            else {
                amt = 1 - (a - 0.9) / 0.1;
            }
            ctx.globalAlpha = amt;
            this.game.mainFont.drawCenteredText(ctx, this.specialMessageText, 320, 30);
            ctx.globalAlpha = 1;
        }

        // Draw victory/defeat message
        let endingTime = this.defeatTime || this.victoryTime;
        if (endingTime !== null) {
            let a = (this.playTime - endingTime) / ENDING_DURATION;
            let amt: number;
            if (a < 0.1) {
                amt = a / 0.1;
            }
            else {
                amt = 1;
            }
            ctx.globalAlpha = amt;
            if (this.defeatTime !== null) {
                this.game.mainFont.drawCenteredText(ctx, "DEFEATED", 320, 170);
            }
            else {
                this.game.mainFont.drawCenteredText(ctx, "VICTORY", 320, 170);
            }
            ctx.globalAlpha = 1;
        }

        // Draw level up message
        if (this.levelUpTime !== null) {
            let a = this.getLevelUpEffectAmount();
            ctx.globalAlpha = a / 6;
            ctx.fillStyle = "#000000";
            ctx.fillRect(0, 0, 640, 360);
            ctx.globalAlpha = a;
            let ls = this.levelUpMessage;
            for (let k = 0; k < ls.length; k++) {
                let line = ls[k];
                this.game.mainFont.drawCenteredText(ctx, line, 320, 44 + k * 20);
            }
            ctx.globalAlpha = 1;
        }
    }

    handleKeyDown(e: KeyboardEvent) {
        if (this.pc.hp <= 0) return true;
        let code = getKeyboardEventCode(e);
        if(code === null) return true;
        if (this.keysDown.has(code)) {
            return true;
        }
        this.keysDown.add(code);
        if (code == "ArrowUp") {
            this.pc.movingUp = true;
            return false;
        }
        else if (code == "ArrowDown") {
            this.pc.movingDown = true;
            return false;
        }
        else if (code == "ArrowLeft") {
            this.pc.movingLeft = true;
            return false;
        }
        else if (code == "ArrowRight") {
            this.pc.movingRight = true;
            return false;
        }
        else if (this.curSkillBindings.hasOwnProperty(code)) {
            this.curSkillBindings[code].handleKeyDown();
            return false;
        }
        return true;
    }

    handleKeyUp(e: KeyboardEvent) {
        if (this.pc.hp <= 0) return true;
        let code = getKeyboardEventCode(e);
        if(code === null) return true;
        if (this.keysDown.has(code)) {
            this.keysDown.delete(code);
        }
        if (code == "ArrowUp") {
            this.pc.movingUp = false;
            return false;
        }
        else if (code == "ArrowDown") {
            this.pc.movingDown = false;
            return false;
        }
        else if (code == "ArrowLeft") {
            this.pc.movingLeft = false;
            return false;
        }
        else if (code == "ArrowRight") {
            this.pc.movingRight = false;
            return false;
        }
        else if (this.curSkillBindings.hasOwnProperty(code)) {
            this.curSkillBindings[code].handleKeyUp();
            return false;
        }
        return true;
    }

}
