let CANNON_DISCLAIMER = "This skill uses the main cannon, and cannot be fired at the same time as other cannon skills.";

abstract class Skill {
    name: string = "";
    message: string = "";
    passive = false;
    sl: ShmupLayer | null = null;
    start(sl: ShmupLayer) {
        this.sl = sl;
    }

    handleKeyDown(): void { };
    handleKeyUp(): void { };
    update(deltat: number): void { };
}

namespace Skills {

    abstract class CannonSkill extends Skill {
        firePeriod: number;

        constructor(firePeriod: number) {
            super();
            this.firePeriod = firePeriod;
        }

        abstract shoot(timeOffset: number): void;

        handleKeyDown() {
            let sl = this.sl!;
            sl.nextCannonShotTime = sl.playTime;
            sl.curCannonAttacks.push(this);
        }

        handleKeyUp() {
            let sl = this.sl!;
            sl.curCannonAttacks = sl.curCannonAttacks.filter(a => a !== this);
        }

        update() {
            let sl = this.sl!;
            let cca = sl.curCannonAttacks;
            if (cca.length > 0 && cca[cca.length - 1] === this) {
                while (sl.playTime >= sl.nextCannonShotTime) {
                    this.shoot(sl.nextCannonShotTime - sl.playTime);
                    sl.nextCannonShotTime += this.firePeriod;
                }
            }
        }

    }

    export class MachineCannon extends CannonSkill {
        sl: ShmupLayer | null = null;
        shotSpeed = 900;
        damage = 1.0;
        name = "Machine Cannon";


        constructor() {
            super(0.01);
            
            let dps = this.damage/this.firePeriod;
            this.message = `An accurate rapid-fire cannon. Physical damage: ${dps}/s. ${CANNON_DISCLAIMER}`;
        }

        shoot(timeOffset: number) {
            let sl = this.sl!
            let pc = sl.pc;
            let x = pc.pos[0] + pc.rad - timeOffset * this.shotSpeed;
            let y = pc.pos[1] + 6 * Math.cos(sl.nextCannonShotTime * 17 * 2 * Math.PI);
            this.sl!.projectiles.push(new Projectile([x, y],
                [this.shotSpeed, 0], [Collision.Physical], this.damage, pc));

            
            sl.cannonTing();
        }
    }

    export class SprayCannon extends CannonSkill {
        sl: ShmupLayer | null = null;
        shotSpeed = 900;
        damage = 0.9;
        name = "Spray Cannon";

        constructor() {
            super(0.01);
            
            let dps = this.damage/this.firePeriod;
            this.message = `A wide-spray rapid-fire cannon. Physical damage: ${dps}/s. ${CANNON_DISCLAIMER}`;
        }

        shoot(timeOffset: number) {
            let sl = this.sl!
            let pc = sl.pc;
            let x = pc.pos[0] + pc.rad - timeOffset * this.shotSpeed;
            let yMod = (sl.nextCannonShotTime * 17) % 1;
            if (yMod > 0.5) yMod = 1 - yMod;
            yMod = yMod - 0.25;
            let y = pc.pos[1] + 6 * yMod;
            this.sl!.projectiles.push(new Projectile([x, y],
                [this.shotSpeed, this.shotSpeed * yMod], [Collision.Physical],
                this.damage, pc));

            sl.cannonTing();
        }
    }

    export class QuantumCannon extends CannonSkill {
        sl: ShmupLayer | null = null;
        shotSpeed = 900;
        damage = 2.5;
        name = "Quantum Cannon";

        constructor() {
            super(0.05);
            
            let dps = 2 * this.damage/this.firePeriod;
            this.message = `An accurate cannon that can attack quantum ghosts. Quantum damage: ${dps}/s. ${CANNON_DISCLAIMER}`;
        }

        shoot(timeOffset: number) {
            let sl = this.sl!
            let pc = sl.pc;
            let x = pc.pos[0] + pc.rad - timeOffset * this.shotSpeed;
            this.sl!.projectiles.push(new Projectile([x, pc.pos[1] + 3],
                [this.shotSpeed, 0], [Collision.Ghost], this.damage, pc));
            this.sl!.projectiles.push(new Projectile([x, pc.pos[1] - 3],
                [this.shotSpeed, 0], [Collision.Ghost], this.damage, pc));
            sl.cannonTing();
        }
    }

    export class HybridCannon extends CannonSkill {
        sl: ShmupLayer | null = null;
        shotSpeed = 900;
        damage = 1.5;
        manaCostPerSecond = 4;
        name = "Hybrid Cannon";

        constructor() {
            super(0.02);
            
            let dps = this.damage/this.firePeriod;
            this.message = `A cannon that deals ${dps}/s physical damage and ${dps}/s quantum damage. Costs ${this.manaCostPerSecond} mana/s. ${CANNON_DISCLAIMER}`;
        }

        shoot(timeOffset: number) {
            let manaCost = this.manaCostPerSecond * this.firePeriod;
            let sl = this.sl!
            if (!sl.trySpendMana(manaCost)) return;
            let pc = sl.pc;
            let x = pc.pos[0] + pc.rad - timeOffset * this.shotSpeed;
            let yMod = (sl.nextCannonShotTime * 17) % 1;
            if (yMod > 0.5) yMod = 1 - yMod;
            yMod = yMod - 0.25;
            this.sl!.projectiles.push(new Projectile([x, pc.pos[1] + 3 * yMod],
                [this.shotSpeed, this.shotSpeed * yMod * 0.3],
                [Collision.Physical], this.damage, pc));
            this.sl!.projectiles.push(new Projectile([x, pc.pos[1] - 3 * yMod],
                [this.shotSpeed, -this.shotSpeed * yMod * 0.3],
                [Collision.Ghost], this.damage, pc));
            sl.cannonTing();
        }
    }

    export class PowerBlast extends Skill {
        shotSpeed = 900;
        damage = 12;
        bulletCount = 40;
        manaCost = 30;
        name = "Power Blast";

        constructor() {
            super();
            let totalDamage = this.damage * this.bulletCount;
            this.message = `Deals ${totalDamage} physical damage in a short burst aimed forward. Costs ${this.manaCost} mana.`;
        }

        handleKeyDown() {
            if (this.sl!.trySpendMana(this.manaCost)) {
                this.shoot();
            }
        }

        shootStep(timeOffset: number, progress: number) {
            let sl = this.sl!;
            let pc = sl.pc;
            let ang = Math.random() * 2 * Math.PI;
            let r = Math.random();
            r = 15 * r * r;
            let x = pc.pos[0] + 4 + r * Math.cos(ang) - timeOffset * this.shotSpeed;
            let y = pc.pos[1] - 4 - (1 - progress * progress) * r * Math.sin(ang);
            this.sl!.projectiles.push(new Projectile([x, y],
                [this.shotSpeed, 0], [Collision.Physical], this.damage, pc));
        }

        async shoot() {
            let sl = this.sl!;
            let t = sl.playTime;
            sl.game.soundPlayer.playSound(sl.game.data['heavyweapon'], false);
            for (let k = 0; k < this.bulletCount; k++) {
                let progress = (k + 1) / this.bulletCount;
                this.shootStep(t - sl.playTime, progress);
                let tStep = Math.random() * 0.005 * progress * progress;
                t += tStep;
                if (t > sl.playTime) {
                    await sl.waitUntilTime(t);
                }
            }
        }
    }

    export class ParticleBomb extends Skill {
        shotSpeed = 900;
        damage = 8;
        bulletCount = 100;
        manaCost = 40;
        name = "Particle Bomb";

        constructor() {
            super();
            let totalDamage = this.damage * this.bulletCount;
            this.message = `An attack that sprays out in a spiral, dealing a total of ${totalDamage} damage that hits both physical enemies and quantum ghosts. Costs ${this.manaCost} mana.`
        }

        handleKeyDown() {
            if (this.sl!.trySpendMana(this.manaCost)) {
                this.shoot();
            }
        }

        shootStep(timeOffset: number, k: number) {
            let sl = this.sl!;
            let pc = sl.pc;
            let ang = k * 2 * Math.PI / 5.12;
            let dx = Math.cos(ang);
            let dy = Math.sin(ang);
            let r = (4 - this.shotSpeed * timeOffset);
            let x = pc.pos[0] + r * dx;
            let y = pc.pos[1] - r * dy;
            let vx = this.shotSpeed * dx;
            let vy = -this.shotSpeed * dy;
            this.sl!.projectiles.push(new Projectile([x, y],
                [vx, vy], [Collision.Physical, Collision.Ghost], this.damage, pc));
        }

        async shoot() {
            let sl = this.sl!;
            let t = sl.playTime;
            for (let k = 0; k < this.bulletCount; k++) {
                this.shootStep(t - sl.playTime, k);
                t += 0.005;
                if (t > sl.playTime) {
                    await sl.waitUntilTime(t);
                }
                sl.game.soundPlayer.playSound(sl.game.data['cannon'], false);
            }
        }
    }

    export class BigBomb extends Skill {
        damage = 400;
        manaCost = 80;
        name = "Rather Big Bomb";

        constructor() {
            super();
            this.message = `A bomb that deals ${this.damage} damage to all enemies on screen, and destroys all projectiles. Costs ${this.manaCost} mana.`;
        }

        handleKeyDown() {
            if (this.sl!.trySpendMana(this.manaCost)) {
                this.shoot();
            }
        }

        async shoot() {
            let sl = this.sl!;
            sl.game.soundPlayer.playSound(sl.game.data['heavyweapon'], false);
            for (let e of sl.enemies) {
                e.takeDamage(this.damage);
            }
            for (let p of sl.projectiles) {
                p.keep = false;
            }
            sl.flashTime = sl.playTime;
        }
    }

    export class GhostBomb extends BigBomb {
        damage = 200;
        manaCost = 35;
        name = "Quantum Bomb";

        constructor() {
            super();
            this.message = `A bomb that deals ${this.damage} damage to ghosts, and destroys their projectiles. Costs ${this.manaCost} mana.`;
        }

        async shoot() {
            let sl = this.sl!;
            sl.game.soundPlayer.playSound(sl.game.data['heavyweapon'], false);
            for (let e of sl.enemies) {
                if (e.ghost) {
                    e.takeDamage(this.damage);
                }
            }
            for (let p of sl.projectiles) {
                if (p.owner.ghost) {
                    p.keep = false;
                }
            }
            sl.flashTime = sl.playTime;
        }
    }

    export class Stab extends Skill {
        damage = 50;
        manaCost = 20;
        dist = 50;
        rad = 40;
        name = "Stab";

        constructor() {
            super()
            this.message = `Teleports a short distance forward, dealing ${this.damage} damage to all enemies around you both physical and quantum, and providing you with a brief moment of invulnerability. Costs ${this.manaCost} mana.`;
        }

        handleKeyDown() {
            let sl = this.sl!;
            if (sl.trySpendMana(this.manaCost)) {
                let [l, t, r, b] = sl.getGameArea();
                let pc = sl.pc;
                let newX = Math.min(r, pc.pos[0] + this.dist);
                pc.moveTo([newX, pc.pos[1]], true);
                sl.damageEnemiesInRadius(pc.pos, this.rad, this.damage);
                sl.makePlayerInvincible(0.5);
                sl.damageCircles.push(
                    new DamageCircle(pc.pos, this.rad, sl.playTime, sl.playTime + 0.15));

                sl.game.soundPlayer.playSound(sl.game.data['stab'], false);
            }
        }
    }

    export class Shield extends Skill {
        manaCost = 50;
        name = "Shield";

        constructor() {
            super();
            this.message = `Shields you from attacks for a moment. Costs ${this.manaCost} mana.`;
        }

        handleKeyDown() {
            let sl = this.sl!;
            if (sl.trySpendMana(this.manaCost)) {
                sl.makePlayerInvincible(3.0);
                sl.game.soundPlayer.playSound(sl.game.data['supportskill'], false);
            }
        }
    }

    export class TimeDilation extends Skill {
        manaCostPerSecond = 25;
        on = false;
        playedStartSound = false;
        amt = 0.25;
        name = "Time Dilation";

        constructor() {
            super();
            this.message = `Slows down time while held. Costs ${this.manaCostPerSecond} mana per (undilated) second.`;
        }

        start(sl: ShmupLayer) {
            this.sl = sl;
            this.on = false;
            this.playedStartSound = false;
        }

        handleKeyDown() {
            this.on = true;
            this.playedStartSound = false;
        }

        handleKeyUp() {
            this.on = false;
        }

        update(deltat: number) {
            let sl = this.sl!;
            let doIt = false;
            if (this.on) {
                let cost = this.manaCostPerSecond * deltat;
                doIt = sl.trySpendMana(cost);
            }

            if (doIt) {
                sl.skillTimeDilation = this.amt;
                if(!this.playedStartSound) {
                    this.playedStartSound = true;
                    sl.game.soundPlayer.playSound(sl.game.data['tdstart'], false);
                }
            } 
            else {
                sl.skillTimeDilation = 1.0;
                if(this.on) {
                    this.on = false;
                    sl.game.soundPlayer.playSound(sl.game.data['tdend'], false);
                }
            } 
        }
    }

    export class Heal extends Skill {
        manaCost = 90;
        name = "Heal";

        constructor() {
            super();
            this.message = `Heals by 1HP. Costs ${this.manaCost} mana.`
        }

        handleKeyDown() {
            let sl = this.sl!;
            let pc = sl.pc;
            if (pc.hp < pc.maxHP) {
                if (sl.trySpendMana(this.manaCost)) {
                    pc.hp += 1;
                    sl.game.soundPlayer.playSound(sl.game.data['heal'], false);
                }
            }
        }
    }

    export class Deflect extends Skill {
        manaCost = 60;
        rad = 100;
        name = "Deflect";

        constructor() {
            super();
            this.message = `Deflects nearby projectiles back at the attackers. Costs ${this.manaCost} mana.`
        }

        handleKeyDown() {
            let sl = this.sl!;
            let pc = sl.pc;
            if (sl.trySpendMana(this.manaCost)) {
                sl.deflectProjectiles(pc.pos, this.rad);
                sl.damageCircles.push(
                    new DamageCircle(pc.pos, this.rad, sl.playTime, sl.playTime + 0.15));

                sl.game.soundPlayer.playSound(sl.game.data['supportskill'], false);
            }
        }
    }

    export class SpiritBoost extends Skill {
        bonus = 0.50;
        name = "Spirit Boost";
        passive = true;

        constructor() {
            super();
            this.message = `Recover mana ${(this.bonus * 100).toFixed(0)}% faster (passive skill).`
        }

        start(sl: ShmupLayer) {
            this.sl = sl;
            sl.manaBonus += this.bonus;
        }
    }

    export class AbsorbSpirit extends Skill {
        manaGain = 2;
        name = "Absorb Spirit";
        passive = true;

        constructor() {
            super();
            this.message = `For every enemy you destroy, you gain ${this.manaGain} mana (passive skill).`
        }

        start(sl: ShmupLayer) {
            this.sl = sl;
            sl.enemyDestroyedCallbacks.push(() => {
                sl.gainMana(this.manaGain);
            });
        }
    }

    export class AbsorbMind extends Skill {
        bonus = 0.50;
        name = "Absorb Mind";
        passive = true;

        constructor() {
            super();
            this.message = `Earn XP ${(this.bonus * 100).toFixed(0)}% faster (passive skill).`;
        }

        start(sl: ShmupLayer) {
            this.sl = sl;
            sl.xpBonus += this.bonus;
        }
    }
}